/* ====================================================================
 * FTD4Linux Software License, Version 1.0
 *
 * Copyright (c) 2000-2004 The FTD4Linux Development Team.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        FTD4Linux Development Team (http://www.ftd4linux.2y.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "FTD" and "FTD4Linux" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact alguno@ftd4linux.2y.net.
 *
 * 5. Products derived from this software may not be called "FTD4Linux",
 *    nor may "FTD4Linux" appear in their name, without prior written
 *    permission of the FTD4Linux Development Team.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE FTD4LINUX DEVELOPMENT TEAM OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "openftd.h"
#include "ftd_app_core.h"

/**
 * Een eventuele foutmelding van de XML-parser.
 */
static char *xml_parser_errmsg = NULL;

/**
 * De callback-functie voor als er iets fout gaat met het parsen van de XML.
 */
static void
xml_parser_error_func (void *ctx, const char *msg, ...)
{
    va_list args;
    char *orig = xml_parser_errmsg;
    char *tmp;

    va_start(args, msg);

#ifdef __CEGCC__
    // Onder CeGCC kan er op dit punt met de originele code die verder hieronder staat
    // een stack overflow optreden.. Moet nog verder onderzocht worden.. Werk er voorlopig
    // omheen door de foutmelding direct naar de stdio te schrijven.
    vprintf(msg, args);
#else
    tmp = g_strdup_vprintf(msg, args);
    if (orig) {
        xml_parser_errmsg = g_strdup_printf("%s %s", orig, tmp);
        g_free(orig);
        g_free(tmp);
    } else {
        xml_parser_errmsg = tmp;
    }
#endif

    va_end(args);
}

/**
 * Leest de gegevens van een event die door de server werden gegeven en zet deze om naar bruikbare informatie.
 *
 * @arg     tekst       De gegevens zoals deze direct van de FTD servers werden ontvangen.
 * @arg     errmsg      Eventuele foutmeldingen zullen hierin bewaard worden
 *
 * @return  Een nieuw event met daarin de gegevens zoals die uit de gegeven tekst zijn gehaald.
 */
FtdEventExtra *
ftd_eventinfo_parse(FtdString *tekst, char **errmsg)
{
    FtdEventExtra *event = NULL;
    xmlDocPtr doc;
    xmlNodePtr cur;

    g_assert(tekst);

    xmlSetGenericErrorFunc(NULL, xml_parser_error_func);

    if (xml_parser_errmsg) {
        g_free(xml_parser_errmsg);
        xml_parser_errmsg = NULL;
    }

    doc = xmlParseMemory(ftd_string_value(tekst), ftd_string_length(tekst));
    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        if (errmsg) {
            *errmsg = g_strdup(xml_parser_errmsg);
        }

        xmlFreeDoc(doc);
        return NULL;
    }

    if (strcasecmp(cur->name, "PostInfo")) {
        if (errmsg) {
            *errmsg = g_strdup("Document of the wrong type, root node != PostInfo");
        }

        xmlFreeDoc(doc);
        return NULL;
    }

    cur = cur->xmlChildrenNode;
    if (!cur) {
        if (errmsg) {
            *errmsg = g_strdup("ftd_eventinfo_parse(): no children of root-node!");
        }

        xmlFreeDoc(doc);
        return NULL;
    }

    while (cur) {
        if (!strcasecmp(cur->name, "Post")) {
            event = ftd_events_parse_event_extra(doc, cur->xmlChildrenNode);
        }
        if (!strcasecmp(cur->name, "MP3")) {
            if (!event) {
                g_warning("ftd_eventinfo_parse(): event == NULL\n");
                cur = cur->next;
                continue;
            }
            event->mp3Tracks = ftd_events_parse_mp3_tracks(doc, cur->xmlChildrenNode);
        }

        cur = cur->next;
    }
    xmlFreeDoc(doc);

    return event;
}

/**
 * Leest de gegevens voor het vinden van een event zoals die direct van de FTD servers werden ontvangen en maakt deze bruikbaar.
 *
 * @arg     buffer      De gegevens zoals die direct van de FTD servers gekregen zijn.
 * @arg     errmsg      Eventuele foutmeldingen worden hierin opgeslagen
 *
 * @return  Een nieuw event met de gegevens uit de buffer in een bruikbare vorm.
 */
FtdEvent *
ftd_eventinfo_parse_findpost(FtdString *buffer, char **errmsg)
{
    xmlDocPtr doc;
    xmlNodePtr cur;

    FtdEvent *event = NULL;

    g_assert(core);
    g_assert(buffer);

    doc = xmlParseMemory(ftd_string_value(buffer), ftd_string_length(buffer));
    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        if (errmsg) {
            *errmsg = g_strdup(xml_parser_errmsg);
        }

        xmlFreeDoc(doc);
        return NULL;
    }

    if (strcasecmp(cur->name, "Post")) {
        if (errmsg) {
            *errmsg = g_strdup("Document of the wrong type, root node != Post");
        }

        xmlFreeDoc(doc);
        return NULL;
    }

    event = ftd_events_parse_event(doc, cur->xmlChildrenNode);

    ftd_category_update(event);

    xmlFreeDoc(doc);

    return event;
}
