 /* ====================================================================
 * FTD4Linux Software License, Version 1.0
 *
 * Copyright (c) 2000-2004 The FTD4Linux Development Team.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        FTD4Linux Development Team (http://www.ftd4linux.2y.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "FTD" and "FTD4Linux" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact alguno@ftd4linux.2y.net.
 *
 * 5. Products derived from this software may not be called "FTD4Linux",
 *    nor may "FTD4Linux" appear in their name, without prior written
 *    permission of the FTD4Linux Development Team.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE FTD4LINUX DEVELOPMENT TEAM OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <zlib.h>

#include "openftd.h"
#include "ftd_app_core.h"
#include "ftd_config.h"
#include "ftd_events.h"
#include "ftd_groups.h"

static gboolean update_busy = FALSE;

/**
/**
 * Haalt de update-bestanden van FTD-server binnen.
 * Let op: deze functie haalt alleen de namen van de bestanden binnen, niet de bestanden zelf.
 *
 * @arg     list_avail_chunks       Deze lijst met ::FtdString pointers naar de namen van de update-bestanden wordt aangevuld.
 * @arg     list_avail_bigchunks    Deze lijst met ::FtdString pointers naar de namen van de update-bestanden wordt aangevuld.
 * @arg     lastCheck               Timestamp van de laatste keer dat er een succesvolle update is gedaan.
 * @arg     newLastCheck            Hierin komt de timestamp van de 'nieuwe' laatste succesvolle update in te staan.
 * @arg     errmsg                  Een eventueel foutbericht van de server wordt hier in geplaatst.
 *
 * @return  Een fout-code die aangeeft of het binnenhalen van de bestandsnamen succesvol was.
 */
FtdErrorCode
ftd_events_fetch_get_available(FtdList **list_avail_chunks, FtdList **list_avail_bigchunks, time_t lastCheck, time_t *newLastCheck, char **errmsg)
{
    FtdList         *params = NULL;
    char            *strLastCheck;
    FtdErrorCode     retval;
    FtdString       *uitvoer;

    strLastCheck = g_strdup_printf("%d", (int) lastCheck);

    params = ftd_list_append(params, ftd_http_parameter_new("lc", strLastCheck));
    g_free(strLastCheck);

    retval = ftd_http_curl_execute(FTD_URL_EVENTS, params, TRUE, &uitvoer, TRUE, NULL, TRUE);

    ftd_http_parameter_free_all(params);

    if (retval != ERROR_NONE) {
        if (retval == ERROR_MSG) {
            *errmsg = g_strdup(ftd_string_value(uitvoer));
        } else {
            *errmsg = NULL;
        }

        return retval;
    }

    /* Ontleedt de XML data om zo de individuele update files te achterhalen */
    retval = ftd_events_list_parse (uitvoer, newLastCheck, list_avail_chunks, list_avail_bigchunks, errmsg);
    ftd_string_free (uitvoer);

    return retval;
}

struct _parse_data {
    GMutex *mutex;          // Deze mutex is nodig zodat er van binnenuit fouten naar buiten gemeld kunnen worden
    int num_new_events;
    FtdEventParseCallback callback;
    gpointer callback_data;

    gboolean error_has_occured;
    FtdErrorCode retval;
    char **errmsg;
};

struct _fetch_data {
    GMutex *mutex;
    GThreadPool *poolParse;
    FtdEventFileCallback callback;
    struct _parse_data *parse_data;
    int file_teller;
    int num_avail_packets;
};

struct _fetch_callback_data {
    int num_avail_packets;
    int file_teller;
    FtdEventFileCallback callback;
};

static gboolean
fetch_callback(gpointer data)
{
    struct _fetch_callback_data *fetch_data = (struct _fetch_callback_data *) data;

    g_assert(fetch_data->callback);
    fetch_data->callback(fetch_data->num_avail_packets, fetch_data->file_teller);

    g_free(fetch_data);

    return FALSE;
}

static void
fetch_func(gpointer data, gpointer user_data, gboolean bigchunk)
{
    struct _fetch_data *fetch_data = (struct _fetch_data *) user_data;
    FtdList *tmplist = NULL;
    char *url;
    char *filename = NULL;
    gzFile gz_fp;
    unsigned char buf[1024];
    int len;
    FtdString *decompressedData;
    FtdString *decompressedData2;
    FtdErrorCode retval;
    FtdErrorCode parseRetval;
    FtdString *uitvoer = NULL;
    gboolean error_has_occured;
    char *errmsg;

    // Controleer of er bij het ophalen van een vorig pakketje een fout opgetreden is
    g_mutex_lock(fetch_data->parse_data->mutex);
    error_has_occured = fetch_data->parse_data->error_has_occured;
    g_mutex_unlock(fetch_data->parse_data->mutex);

    if (error_has_occured) {
g_print("error has occured!\n");
        return;
    }

    if (bigchunk) {
        // We hebben een bigchunk pakketje
        url = g_strdup_printf("%s%s.gz", FTD_BIGDUMPS_DIR, (char *) data);
    } else {
        url = g_strdup_printf("%s%s.gz", FTD_DUMPS_DIR, (char *) data);
    }

    retval = ftd_http_curl_execute(url, NULL, TRUE, &uitvoer, TRUE, &filename, TRUE);

    if (retval != ERROR_NONE) {
        g_mutex_lock(fetch_data->parse_data->mutex);
        if (!fetch_data->parse_data->error_has_occured) {
            fetch_data->parse_data->error_has_occured = TRUE;
            fetch_data->parse_data->retval = retval;
            if (uitvoer) {
                *fetch_data->parse_data->errmsg = g_strdup(ftd_string_value(uitvoer));
                ftd_string_free(uitvoer);
            }
        }
        g_mutex_unlock(fetch_data->parse_data->mutex);

        return;
    }

    // Decodeer de gzip data
    gz_fp = gzopen(filename, "rb");

    decompressedData = ftd_string_new("");
    memset(buf, 0, sizeof(buf));
    while ((len = gzread(gz_fp, buf, sizeof(buf) - 1)) > 0) {
        decompressedData = ftd_string_append(decompressedData, buf);
        memset(buf, 0, sizeof(buf));
    }

    gzclose(gz_fp);
    unlink(filename);

    if (len == -1) {
        // Hiermee kunnen we de GUI signaleren dat er tijdens
        // het lezen van de server iets misgegaan is
        g_mutex_lock(fetch_data->parse_data->mutex);
        if (!fetch_data->parse_data->error_has_occured) {
            fetch_data->parse_data->error_has_occured = TRUE;
            fetch_data->parse_data->retval = ERROR_UNKNOWN;
            *fetch_data->parse_data->errmsg = g_strdup_printf("Error while decoding file %s", filename);
        }
        g_mutex_unlock(fetch_data->parse_data->mutex);

        g_free(filename);

        return;
    }

    g_free(filename);

    // Filter evt. vage tekens eruit omdat de XML parser anders gaat mekkeren
    decompressedData2 = ftd_string_replace2(ftd_string_value(decompressedData), "", "'");

    parseRetval = ftd_events_parse (decompressedData2, NULL, &tmplist, &errmsg, FALSE);
    ftd_string_free(decompressedData2);
    ftd_string_free(decompressedData);

    if (parseRetval != ERROR_NONE) {
        g_mutex_lock(fetch_data->parse_data->mutex);
        if (!fetch_data->parse_data->error_has_occured) {
            fetch_data->parse_data->error_has_occured = TRUE;
            fetch_data->parse_data->retval = parseRetval;
            if (errmsg) {
                *fetch_data->parse_data->errmsg = g_strdup(errmsg);
                g_free(errmsg);
            }
        }
        g_mutex_unlock(fetch_data->parse_data->mutex);

        return;
    }

    g_mutex_lock(fetch_data->mutex);
    fetch_data->file_teller++;
    g_mutex_unlock(fetch_data->mutex);

    g_free(url);

    if (fetch_data->callback) {
        struct _fetch_callback_data *callback_data = g_malloc0(sizeof(struct _fetch_callback_data));

        callback_data->callback = fetch_data->callback;
        callback_data->num_avail_packets = fetch_data->num_avail_packets;
        callback_data->file_teller = fetch_data->file_teller;

        g_idle_add(fetch_callback, callback_data);
    }

    g_thread_pool_push(fetch_data->poolParse, tmplist, NULL);

    return;
}

static void
fetch_func_normal(gpointer data, gpointer user_data)
{
    fetch_func(data, user_data, FALSE);
}

static void
fetch_func_bigchunk(gpointer data, gpointer user_data)
{
    fetch_func(data, user_data, TRUE);
}

static void
parse_func(gpointer data, gpointer user_data)
{
    struct _parse_data *parse_data = (struct _parse_data *) user_data;
    FtdList *eventlistNieuw = (FtdList *) data;
    FtdList *list;

    parse_data->num_new_events += ftd_list_length(eventlistNieuw);

    /* Draai de eventlistNieuw om zodat de meest nieuwe events als eerste behandeld worden */
    eventlistNieuw = ftd_list_reverse(eventlistNieuw);

    /* Bewaar alle nieuw opgehaalde postings direct */
    ftd_eventlist_merge(core->eventlist, eventlistNieuw, parse_data->callback, parse_data->callback_data);

    /* Geef de elementen uit eventlistNieuw weer vrij */
    list = eventlistNieuw;
    while (list) {
        FTDEventObject *event = ftd_list_value(list);
        ftd_event_object_destroy(event);

        list = ftd_list_next(list);
    }

    return;
}

/**
 * Haalt alle updates binnen die binnen gehaald kunnen of mogen worden.
 *
 * @arg     file_callback   Een callback om de voortgang bij te houden van het downloaden.
 * @arg     parse_callback  Een callback om de voortgang bij te houden van het parsen van de gegevens.
 * @arg     status_callback Een callback om de algemene voortgang bij te houden.
 * @arg     errmsg          Een eventueel foutbericht van de FTD server komt hier in te staan.
 * @arg     data            Callback-data
 *
 * @return  Een fout-code die aangeeft of het updaten succesvol was of niet.
 */
FtdErrorCode
ftd_events_fetch(FtdEventFileCallback file_callback, FtdEventParseCallback parse_callback, FtdEventStatusCallback status_callback, char **errmsg, gpointer data)
{
    int              retval;
    FtdList         *listAvailChunks = NULL;
    FtdList         *listAvailBigChunks = NULL;
    FtdList         *eventlistNieuw = NULL;
    FtdString       *uitvoer;
    int              origPrevCheck;
    time_t           newLastCheck;
    FtdErrorCode     parseRetval;
    FtdList         *list;
    int              file_teller;

    g_assert (core);
    g_assert (core->config);

    /* Ruim (te) oude FTD postings op */
    ftd_eventlist_cleanup(core->eventlist);

    if (file_callback) {
        file_callback(0, 0);
    }

    retval = ftd_events_fetch_get_available(&listAvailChunks, &listAvailBigChunks, core->config->lastCheck, &newLastCheck, errmsg);

    if (retval != ERROR_NONE) {
        return retval;
    }

    // Er kunnen nieuwe groepen zojuist toegevoegd zijn aan de lijst
    ftd_group_save();

    ftd_categories_reset_counters();

    // Behandel eerste de kleine chunks omdat deze meer recent zijn dan de bigchunks
    list = listAvailChunks;
    file_teller = 0;

    while (list) {

        FtdList *tmplist = NULL;
        char *url = g_strdup_printf("%s%s.gz", FTD_DUMPS_DIR, (char *) ftd_list_value(list));
        char *filename = NULL;
        gzFile gz_fp;
        unsigned char buf[1024];
        int len;
        FtdString *decompressedData;
        FtdString *decompressedData2;

        file_teller++;

        if (file_callback) {
            file_callback(ftd_list_length(listAvailChunks) + ftd_list_length(listAvailBigChunks), file_teller);
        }

        retval = ftd_http_curl_execute(url, NULL, TRUE, NULL, TRUE, &filename, TRUE);
        g_free(url);

        if (retval != ERROR_NONE) {
            return retval;
        }

        // Decodeer de gzip data
        gz_fp = gzopen(filename, "rb");

        decompressedData = ftd_string_new("");
        memset(buf, 0, sizeof(buf));
        while ((len = gzread(gz_fp, buf, sizeof(buf) - 1)) > 0) {
            decompressedData = ftd_string_append(decompressedData, buf);
            memset(buf, 0, sizeof(buf));
        }

        gzclose(gz_fp);
        unlink(filename);
        g_free(filename);

        if (len == -1) {
            // Hiermee kunnen we de GUI signaleren dat er tijdens
            // het lezen van de server iets misgegaan is
            return ERROR_UNKNOWN;
        }

        // Filter evt. vage tekens eruit omdat de XML parser anders gaat mekkeren
        decompressedData2 = ftd_string_replace2(ftd_string_value(decompressedData), "", "'");

        parseRetval = ftd_events_parse (decompressedData2, NULL, &tmplist, errmsg, FALSE);
        ftd_string_free(decompressedData2);
        ftd_string_free(decompressedData);
        uitvoer = NULL;

        if (parseRetval != ERROR_NONE) {
            return parseRetval;
        }

        eventlistNieuw = g_slist_concat(eventlistNieuw, tmplist);

        g_free(ftd_list_value(list));
        list = ftd_list_next(list);
    }

    ftd_list_free(listAvailChunks);

    // Behandel nu de grote chunks een controle op de ftd38Webscripts config optie is niet nodig
    list = listAvailBigChunks;
    file_teller = 0;

    while (list) {
        FtdList *tmplist = NULL;
        char *url = g_strdup_printf("%s%s.gz", FTD_BIGDUMPS_DIR, (char *) ftd_list_value(list));
        char *filename = NULL;
        gzFile gz_fp;
        unsigned char buf[1024];
        int len;
        FtdString *decompressedData;
        FtdString *decompressedData2;

        file_teller++;

        if (file_callback) {
            file_callback(ftd_list_length(listAvailChunks) + ftd_list_length(listAvailBigChunks), file_teller);
        }

        retval = ftd_http_curl_execute(url, NULL, TRUE, NULL, TRUE, &filename, TRUE);
        g_free(url);

        if (retval != ERROR_NONE) {
            return retval;
        }

        // Decodeer de gzip data
        gz_fp = gzopen(filename, "rb");

        decompressedData = ftd_string_new("");
        memset(buf, 0, sizeof(buf));
        while ((len = gzread(gz_fp, buf, sizeof(buf) - 1)) > 0) {
            decompressedData = ftd_string_append(decompressedData, buf);
            memset(buf, 0, sizeof(buf));
        }

        gzclose(gz_fp);
        unlink(filename);
        g_free(filename);

        if (len == -1) {
            // Hiermee kunnen we de GUI signaleren dat er tijdens
            // het lezen van de server iets misgegaan is
            return ERROR_UNKNOWN;
        }

        // Filter evt. vage tekens eruit omdat de XML parser anders gaat mekkeren
        decompressedData2 = ftd_string_replace2(ftd_string_value(decompressedData), "", "'");

        parseRetval = ftd_events_parse (decompressedData2, NULL, &tmplist, errmsg, FALSE);
        ftd_string_free(decompressedData2);
        ftd_string_free(decompressedData);
        uitvoer = NULL;

        if (parseRetval != ERROR_NONE) {
            return parseRetval;
        }

        eventlistNieuw = g_slist_concat(eventlistNieuw, tmplist);

        g_free(ftd_list_value(list));
        list = ftd_list_next(list);
    }

    ftd_list_free(listAvailBigChunks);

    if (eventlistNieuw == NULL) {    /* De postlist kan leeg zijn (dus geen nieuwe postings of status
                                      * veranderingen sinds laatste update) */
        return ERROR_NEW;
    }

    if (core->config->lastCheck != 0) {
        ftd_eventlist_load(core->eventlist, NULL);
    }

    /* Pas in de config file de lastCheck en prevCheck waarden aan */
    core->config->prevCheck = core->config->lastCheck;
    core->config->lastCheck = newLastCheck;
    ftd_config_save();

    /* Draai de eventlistNieuw om zodat de meest nieuwe events als eerste behandeld worden */
    eventlistNieuw = ftd_list_reverse(eventlistNieuw);

    /* Bewaar alle nieuw opgehaalde postings direct */
    ftd_eventlist_merge(core->eventlist, eventlistNieuw, parse_callback, data);

    if (status_callback) {
        status_callback(data);
    }

    ftd_eventlist_save(core->eventlist);

    if (status_callback) {
        status_callback(data);
    }

    return ERROR_NONE;
}

struct _file_callback_data {
    FtdEventFileCallback callback;
    int totaal;
    int huidig;
    gpointer data;
};

static gboolean
launch_file_callback(gpointer data)
{
    struct _file_callback_data *callback_data = (struct _file_callback_data *) data;

    callback_data->callback(callback_data->totaal, callback_data->huidig/*, callback_data->data */);

    g_free(callback_data);

    return FALSE;
}

static void
trigger_file_callback(FtdEventFileCallback callback, int totaal, int huidig, gpointer data)
{
    struct _file_callback_data *callback_data;

    if (!callback) {
        return;
    }

    callback_data = g_malloc0(sizeof(struct _file_callback_data));
    callback_data->callback = callback;
    callback_data->totaal = totaal;
    callback_data->huidig = huidig;
    callback_data->data = data;

    g_idle_add(launch_file_callback, callback_data);
}

struct _status_callback_data {
    FtdEventStatusCallback callback;
    gpointer data;
};

static gboolean
launch_status_callback(gpointer data)
{
    struct _status_callback_data *callback_data = (struct _status_callback_data *) data;

    callback_data->callback(callback_data->data);

    g_free(callback_data);

    return FALSE;
}

static void
trigger_status_callback(FtdEventStatusCallback callback, gpointer data)
{
    struct _status_callback_data *callback_data;

    if (!callback) {
        return;
    }

    callback_data = g_malloc0(sizeof(struct _status_callback_data));
    callback_data->callback = callback;
    callback_data->data = data;

    g_idle_add(launch_status_callback, callback_data);
}

struct _done_callback_data {
    FtdUpdateDoneCallback callback;
    FtdErrorCode retval;
    const char *errmsg;
    gpointer data;
};

static gboolean
launch_done_callback(gpointer data)
{
    struct _done_callback_data *callback_data = (struct _done_callback_data *) data;

    callback_data->callback(callback_data->retval, callback_data->errmsg, callback_data->data);

    if (callback_data->errmsg) {
        g_free((gpointer) callback_data->errmsg);
    }

    g_free(callback_data);

    return FALSE;
}

static void
trigger_done_callback(FtdUpdateDoneCallback callback, FtdErrorCode retval, const char *errmsg, gpointer data)
{
    struct _done_callback_data *callback_data;

    if (!callback) {
        return;
    }

    callback_data = g_malloc0(sizeof(struct _done_callback_data));
    callback_data->callback = callback;
    callback_data->retval = retval;
    if (errmsg) {
        callback_data->errmsg = g_strdup(errmsg);
    }
    callback_data->data = data;

    g_idle_add(launch_done_callback, callback_data);
}

struct _fetch_async_data {
    FtdEventFileCallback file_callback;
    FtdList *listAvailChunks;
    FtdList *listAvailBigChunks;
    FtdEventParseCallback parse_callback;
    FtdEventStatusCallback status_callback;
    FtdUpdateDoneCallback done_callback;
    gpointer data;

    time_t newLastCheck;
};

static gpointer
fetch_async_func(gpointer data)
{
    GThreadPool     *poolDownloadNormal;
    GThreadPool     *poolDownloadBigChunk;
    GThreadPool     *poolParse;
    struct _fetch_data fetch_data;
    struct _parse_data parse_data;
    struct _fetch_async_data *fetch_async_data = (struct _fetch_async_data *) data;
    char            *errmsg = NULL;
    FtdList         *list;

    memset(&fetch_data, 0, sizeof(fetch_data));
    memset(&parse_data, 0, sizeof(parse_data));

    poolDownloadNormal = g_thread_pool_new(fetch_func_normal, &fetch_data, 4, TRUE, NULL);
    poolDownloadBigChunk = g_thread_pool_new(fetch_func_bigchunk, &fetch_data, 4, TRUE, NULL);
    poolParse = g_thread_pool_new(parse_func, &parse_data, 1, TRUE, NULL);

    fetch_data.mutex = g_mutex_new();
    fetch_data.callback = fetch_async_data->file_callback;
    fetch_data.poolParse = poolParse;
    fetch_data.parse_data = &parse_data;
    fetch_data.num_avail_packets = ftd_list_length(fetch_async_data->listAvailChunks) + ftd_list_length(fetch_async_data->listAvailBigChunks);
    parse_data.errmsg = &errmsg;
    parse_data.mutex = g_mutex_new();
    parse_data.callback = fetch_async_data->parse_callback;
    parse_data.callback_data = fetch_async_data->data;

    // Begin met de meest recente kleine chunk-pakketten
    list = ftd_list_reverse(fetch_async_data->listAvailChunks);

    if (core->config->lastCheck != 0) {
        ftd_eventlist_load(core->eventlist, NULL);
    }

    while (list) {
        g_thread_pool_push(poolDownloadNormal, ftd_list_value(list), NULL);
        list = ftd_list_next(list);
    }

    // Behandel nu de grote chunk-pakketten
    list = ftd_list_reverse(fetch_async_data->listAvailBigChunks);

    while (list) {
        g_thread_pool_push(poolDownloadBigChunk, ftd_list_value(list), NULL);
        list = ftd_list_next(list);
    }

    // Wacht totdat alles gedownload is
    g_thread_pool_free(poolDownloadNormal, FALSE, TRUE);
    g_thread_pool_free(poolDownloadBigChunk, FALSE, TRUE);

    // Wacht totdat alles geparsed is
    g_thread_pool_free(poolParse, FALSE, TRUE);

    // Geef alle elementen uit de chunks-lijsten vrij
    list = fetch_async_data->listAvailChunks;
    while (list) {
        g_free(ftd_list_value(list));
        list = ftd_list_next(list);
    }

    list = fetch_async_data->listAvailBigChunks;
    while (list) {
        g_free(ftd_list_value(list));
        list = ftd_list_next(list);
    }

    ftd_list_free(fetch_async_data->listAvailChunks);
    ftd_list_free(fetch_async_data->listAvailBigChunks);

    g_mutex_free(parse_data.mutex);

    if (parse_data.error_has_occured) {
        update_busy = FALSE;

        trigger_done_callback(fetch_async_data->done_callback, parse_data.retval, *parse_data.errmsg, fetch_async_data->data);

        if (parse_data.errmsg) {
            g_free(*parse_data.errmsg);
        }

        g_free(fetch_async_data);
        return;
    }

    if (parse_data.num_new_events == 0) {    /* De postlist kan leeg zijn (dus geen nieuwe postings of status
                                                * veranderingen sinds laatste update) */
        update_busy = FALSE;

        trigger_done_callback(fetch_async_data->done_callback, ERROR_NEW, NULL, fetch_async_data->data);

        g_free(fetch_async_data);

        return;
    }

    /* Pas in de config file de lastCheck en prevCheck waarden aan */
    core->config->prevCheck = core->config->lastCheck;
    core->config->lastCheck = fetch_async_data->newLastCheck;
    ftd_config_save();

    trigger_status_callback(fetch_async_data->status_callback, data);

    ftd_eventlist_save(core->eventlist);

    trigger_status_callback(fetch_async_data->status_callback, data);

    trigger_done_callback(fetch_async_data->done_callback, ERROR_NONE, NULL, fetch_async_data->data);

    g_free(fetch_async_data);

    update_busy = FALSE;
}

/**
 * Haalt alle updates binnen die binnen gehaald kunnen of mogen worden.
 *
 * @arg     file_callback   Een callback om de voortgang bij te houden van het downloaden.
 * @arg     parse_callback  Een callback om de voortgang bij te houden van het parsen van de gegevens.
 * @arg     status_callback Een callback om de algemene voortgang bij te houden.
 * @arg     done_callback   Een callback om aan te geven dat het updaten voltooid is
 * @arg     data            Callback-data
 *
 * @return  Een fout-code die aangeeft of het updaten succesvol was of niet.
 */
void
ftd_events_fetch_async(FtdEventFileCallback file_callback, FtdEventParseCallback parse_callback, FtdEventStatusCallback status_callback, FtdUpdateDoneCallback done_callback, gpointer data)
{
    int              retval;
    FtdList         *listAvailChunks = NULL;
    FtdList         *listAvailBigChunks = NULL;
    FtdList         *eventlistNieuw = NULL;
    FtdString       *uitvoer;
    int              origPrevCheck;
    time_t           newLastCheck;
    FtdErrorCode     parseRetval;
    char            *errmsg;
    struct _fetch_async_data *fetch_async_data;

    g_assert (core);
    g_assert (core->config);

    if (update_busy) {
        return;
    }

    update_busy = TRUE;

    /* Ruim (te) oude FTD postings op */
    ftd_eventlist_cleanup(core->eventlist);

    if (file_callback) {
        file_callback(0, 0);
    }

    retval = ftd_events_fetch_get_available(&listAvailChunks, &listAvailBigChunks, core->config->lastCheck, &newLastCheck, &errmsg);

    if (retval != ERROR_NONE) {
        if (done_callback) {
            done_callback(retval, errmsg, data);
        }

        if (errmsg) {
            g_free(errmsg);
        }

        update_busy = FALSE;

        return;
    }

    // Er kunnen nieuwe groepen zojuist toegevoegd zijn aan de lijst
    ftd_group_save();

    ftd_categories_reset_counters();

    fetch_async_data = g_malloc0(sizeof(struct _fetch_async_data));
    fetch_async_data->parse_callback = parse_callback;
    fetch_async_data->file_callback = file_callback;
    fetch_async_data->status_callback = status_callback;
    fetch_async_data->done_callback = done_callback;
    fetch_async_data->data = data;
    fetch_async_data->listAvailChunks = listAvailChunks;
    fetch_async_data->listAvailBigChunks = listAvailBigChunks;
    fetch_async_data->newLastCheck = newLastCheck;

    g_thread_create(fetch_async_func, fetch_async_data, FALSE, NULL);
}
