/* ====================================================================
 * FTD4Linux Software License, Version 1.0
 *
 * Copyright (c) 2000-2004 The FTD4Linux Development Team.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        FTD4Linux Development Team (http://www.ftd4linux.2y.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "FTD" and "FTD4Linux" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact alguno@ftd4linux.2y.net.
 *
 * 5. Products derived from this software may not be called "FTD4Linux",
 *    nor may "FTD4Linux" appear in their name, without prior written
 *    permission of the FTD4Linux Development Team.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE FTD4LINUX DEVELOPMENT TEAM OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "openftd.h"
#include "ftd_app_core.h"

/**
 * Vraagt het saldo en maximum aantal te geven kudos voor een event op.
 *
 * @arg     post_id         Het FTD-nummer van het event waarvoor kudos gegeven zouden kunnen worden.
 * @arg     saldo           Hierin wordt het kudo-saldo van de gebruiker gezet.
 * @arg     max_te_geven    Hierin wordt gezet hoeveel kudos de gerbuiker nog aan het gegeven event mag geven.
 * @arg     errmsg          Een eventueel foutbericht van de FTD servers wordt hier in gezet.
 *
 * @return  Een fout-code die aangeeft of het ophalen van de informaite succesvol was.
 */
FtdErrorCode
ftd_kudos_get(int post_id, int *saldo, int *max_te_geven, char **errmsg)
{
    FtdList *params = NULL;
    char tmp[16];
    int retval;
    FtdString *uitvoer;
    g_assert (core);

    snprintf(tmp, sizeof(tmp), "%i", post_id);

    params = ftd_list_append(params, ftd_http_parameter_new("mode", "get"));
    params = ftd_list_append(params, ftd_http_parameter_new("id", tmp));

    retval = ftd_http_curl_execute(FTD_URL_KUDOS, params, TRUE, &uitvoer, TRUE, NULL, TRUE);

    ftd_http_parameter_free_all(params);

    if (retval != ERROR_NONE) {
        if (retval == ERROR_MSG) {
            *errmsg = g_strdup(ftd_string_value(uitvoer));
        } else {
            *errmsg = NULL;
        }

        ftd_string_free(uitvoer);

        return retval;
    }

    // Uitvoer hoort een bericht in de trand van '100*20' te zijn
    // 100 == het saldo van de ftd gebruiker
    // 20 == het aantal kudos dat nog aan de geselecteerde posting gegeven mag worden

    if (!strstr(ftd_string_value(uitvoer), "*")) {
        *errmsg = NULL;
        return ERROR_PARSE;
    }

    *saldo = atoi(ftd_string_value(uitvoer));
    *max_te_geven = atoi(strstr(ftd_string_value(uitvoer), "*") + 1);

    ftd_string_free(uitvoer);
    return ERROR_NONE;
}

/**
 * Geeft opdracht kudos te geven aan een event.
 *
 * @arg     post_id         Het FTD-nummer van het event waar de kudos voor bestemd zijn.
 * @arg     kudos           Het aantal kudos dat gegeven moet worden.
 * @arg     errmsg          Een eventueel foutbericht van de server komt hier in te staan.
 *
 * @return  Een fout-code die aangeeft of het geven van de kudos succesvol was.
 */
FtdErrorCode
ftd_kudos_give(int post_id, int kudos, char **errmsg)
{
    FtdList *params = NULL;
    char tmp[16];
    int retval;
    FtdString *uitvoer;

    g_assert (core);

    params = ftd_list_append(params, ftd_http_parameter_new("mode", "give"));
    snprintf(tmp, sizeof(tmp), "%i", post_id);
    params = ftd_list_append(params, ftd_http_parameter_new("id", tmp));
    snprintf(tmp, sizeof(tmp), "%i", kudos);
    params = ftd_list_append(params, ftd_http_parameter_new("kudos", tmp));

    retval = ftd_http_curl_execute(FTD_URL_KUDOS, params, TRUE, &uitvoer, TRUE, NULL, TRUE);

    ftd_http_parameter_free_all(params);

    if (retval != ERROR_NONE) {
        if (retval == ERROR_MSG) {
            *errmsg = g_strdup(ftd_string_value(uitvoer));
        } else {
            *errmsg = NULL;
        }

        ftd_string_free(uitvoer);

        return retval;
    }

    // Uitvoer hoort een bericht in de trand van '100*20' te ontvangen
    // 100 == het saldo van de ftd gebruiker
    // 20 == het aantal kudos dat nog aan de geselecteerde posting gegeven mag worden

    // We hebben net kudos uitgegeven..we doen er verder niets meer mee
    ftd_string_free(uitvoer);
    return ERROR_NONE;
}
