/* ====================================================================
 * FTD4Linux Software License, Version 1.0
 *
 * Copyright (c) 2000-2004 The FTD4Linux Development Team.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        FTD4Linux Development Team (http://www.ftd4linux.2y.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "FTD" and "FTD4Linux" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact alguno@ftd4linux.2y.net.
 *
 * 5. Products derived from this software may not be called "FTD4Linux",
 *    nor may "FTD4Linux" appear in their name, without prior written
 *    permission of the FTD4Linux Development Team.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE FTD4LINUX DEVELOPMENT TEAM OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>

#include "openftd.h"

/**
 * @file    ftd_splitted_list.c
 *
 * Een splitted list is feitelijk niets dan een ::FtdList met ::FtdString pointers erin. De opslag van deze lijsten is echter bijzonder.
 * Ze worden opgeslagen door alle strings achtereen te zetten en er pipes (|) tussen te zetten. Dit teken mag dan ook niet voorkomen in
 * de strings in deze lijsten.
 */

/**
 * Maakt een nieuwe splitted list aan met dezelfde inhoud als de oude (deep copy).
 *
 * @arg     splitted_list       Een lijst met ::FtdString pointers die gekopieerd moet worden.
 *
 * @return  Een nieuwe lijst met ::FtdString pointers die wijzen naar nieuwe strings met dezelfde inhoud als die in de originele lijst.
 */
FtdList *
ftd_splitted_list_duplicate(FtdList *splitted_list)
{
    FtdList *origlist = splitted_list;
    FtdList *ret = NULL;

    while (origlist) {
        FtdString *item = ftd_list_value(origlist);

        g_assert(item);

        ret = ftd_list_append(ret, ftd_string_new(ftd_string_value(item)));

        origlist = ftd_list_next(origlist);
    }

    return ret;

}

/**
 * Maakt een splitted list van een enkele string.
 *
 * @arg     tekst       De string met alle waarden, gescheiden door pipes (|),
 *
 * @return  Een nieuwe lijst met ::FtdString pointers met daarin de waarden zoals in tekst gevonden.
 */
FtdList *
ftd_splitted_list_split(const char *tekst)
{
    gchar **splitted;
    FtdList *list = NULL;

    int i = 0;
    splitted = g_strsplit(tekst, "|", 0);
    while(splitted[i] != NULL) {
        if (strlen(splitted[i]) > 0) {
            list = ftd_list_append(list, ftd_string_new(splitted[i]));
        }
        i++;
    }
    g_strfreev(splitted);

    return list;
}

/**
 * Maakt een enkele string van een splitted list.
 *
 * @arg     splitted_list       Een lijst van ::FtdString pointers met alle waarden.
 *
 * @return  Een string met daarin alle waarden uit splitted_list aaneen geschakeld, gescheiden door pipes (|).
 */
FtdString *
ftd_splitted_list_concat(FtdList *splitted_list)
{
    gboolean beenhere = FALSE;
    FtdString *res = ftd_string_new("|");
    FtdList *list = splitted_list;

    while (list) {
        FtdString *item = ftd_list_value(list);

        g_assert(item);

        if (!beenhere) {
            beenhere = TRUE;
        } else {
            res = ftd_string_append(res, "|");
        }

        res = ftd_string_append(res, ftd_string_value(item));

        list = ftd_list_next(list);
    }

    // Zowel aan het begin als aan het einde moet een | staan
    if (ftd_string_length(res) > 1) {
        res = ftd_string_append(res, "|");
    }

    return res;
}

/**
 * Geeft een splitted list vrij.
 * De lijst zelf wordt vrijgegeven en NULL gemaakt.
 *
 * @arg     splitted_list       Een lijst met ::FtdString pointers om te vrij te geven.
 */
void
ftd_splitted_list_free(FtdList **splitted_list)
{
    FtdList *list = *splitted_list;

    while (list) {
        FtdString *item = ftd_list_value(list);
        g_assert(item);
        ftd_string_free(item);
        list = ftd_list_next(list);
    }

    ftd_list_free(*splitted_list);
    *splitted_list = NULL;
}
